/*
   TVSPAWN.CPP
   This is a demo for calling programs from within the TV environment
   without destroying the desktop.
   The approach is as follows:
   - overload TApplication::suspend() and TApplication::resume().
     We must do this to avoid TScreen::suspend() and TScreen::resume()
     to be executed.
   - display a non-movable, non-resizable TWindow on the screen
     (see class TSwanWindow)
   - define a window() (see conio.h) that covers the client
     area of our TSpawnWindow and set "directvideo=1"
   - redirect INT 29h "fast console output" to an own interrupt
     routine (INT 29h is called everytime a character is written to
     stdout). This routine calls cprintf() to display the character
   - proceed with spawning a child process as described in
     Borland examples
   - restore INT 29h to the original

   I tested the above on different standard text modes and encountered
   no problems. Nevertheless this example comes without any warranty.

   Please report any experiences/errors/enhancements to
   Frank Hohmann
   University of Osnabrueck, Germany
   e-mail: fhohmann@rols1.oec.uni-osnabrueck.de
*/

#include <stdlib.h>               // for system()
#include <dos.h>                  // for setvect(), getvect()
#include <conio.h>                // cprintf(), window()
#include <stdio.h>

#define Uses_TEvent
#define Uses_TProgram
#define Uses_TApplication
#define Uses_TKeys
#define Uses_TRect
#define Uses_TMenuBar
#define Uses_TSubMenu
#define Uses_TMenuItem
#define Uses_TStatusLine
#define Uses_TStatusItem
#define Uses_TStatusDef
#define Uses_TDeskTop
#define Uses_TWindow
#define Uses_TSystemError
#define Uses_TEventQueue

#include <tv.h>

#define INT_NUMBER 0x29        // interrupt for "fast console output"
#define DOS_CMD    "MEM /C"    // command to spawn for test purpose

const int cmNewSpawnWin   = 100;
const int cmSpawn      = 1000;


class TVApp : public TApplication
{

public:
    TVApp();
    static TStatusLine *initStatusLine( TRect r );
    static TMenuBar *initMenuBar( TRect r );
    virtual void handleEvent( TEvent& event);
    void mySpawnWindow();
    virtual void suspend();
    virtual void resume();
};



class TSpawnWindow : public TWindow
{

public:
   TSpawnWindow( const TRect& r, const char *aTitle, short aNumber );
   virtual void handleEvent(TEvent& event);
   void spawn();
};


// ------------------- TVApp
TVApp::TVApp() :
    TProgInit( &TVApp::initStatusLine,
	       &TVApp::initMenuBar,
	       &TVApp::initDeskTop
	     )
{
}

TStatusLine *TVApp::initStatusLine(TRect r)
{
    r.a.y = r.b.y - 1;     // move top to 1 line above bottom
    return new TStatusLine( r,
	*new TStatusDef( 0, 0xFFFF ) +
	    *new TStatusItem( 0, kbF10, cmMenu ) +
	    *new TStatusItem( 0, kbAltF3, cmClose ) +
	    *new TStatusItem( "~Alt-X~ Exit", kbAltX, cmQuit )
	);
}

TMenuBar *TVApp::initMenuBar( TRect r )
{

    r.b.y = r.a.y + 1;    // set bottom line 1 line below top line
    TMenuBar *m =
    new TMenuBar( r,
	*new TSubMenu( "~S~pawn", kbAltF )+
	    *new TMenuItem( "~T~est",  cmNewSpawnWin,   kbF4, hcNoContext, "F4" )+
	    newLine()+
	    *new TMenuItem( "E~x~it", cmQuit, cmQuit, hcNoContext, "Alt-X" )
	);
   return m;
}

void TVApp::handleEvent(TEvent& event)
{
    TApplication::handleEvent(event);
    if( event.what == evCommand )
	{
	switch( event.message.command )
	    {
	    case cmNewSpawnWin:
		mySpawnWindow();
		break;
	    default:
		return;
	    }
	clearEvent( event );       // clear event after handling
	}
}

void TVApp::mySpawnWindow()
{
    TRect r( 2, 2, 78, 21 );

    // insert a new TSpawnWindow
    TSpawnWindow *window = new TSpawnWindow ( r, "Output", wnNoNumber);
    deskTop->insert(window);
    // ... and let it do its job
    message(window, evBroadcast, cmSpawn, NULL);
}

// We must override TApplication::suspend()
// since we don't want TScreen::suspend() to be executed
void TVApp::suspend()
{
    TSystemError::suspend();
    TEventQueue::suspend();
}

// We must override TApplication::resume()
// since we don't want TScreen::resume() to be executed
void TVApp::resume()
{
    TEventQueue::resume();
    TSystemError::resume();
}


// ------------------- TSpawnWindow
TSpawnWindow::TSpawnWindow( const TRect& r, const char *aTitle, short aNumber):
			  TWindow( r, aTitle, aNumber),
			  TWindowInit( &TSpawnWindow::initFrame)

{
  flags = 0; // no Zoom/close/move style bits
}


#pragma argsused
void interrupt NewInt29( unsigned bp, unsigned di, unsigned si,
		       unsigned ds, unsigned es, unsigned dx,
		       unsigned cx, unsigned bx, unsigned ax,
		       unsigned ip, unsigned cs, unsigned flags )
{
   // the character to be outputted must be taken from AL
   cprintf("%c", (ax &~ 0xFF00));
   // signal end of interrupt (EOI)
   outportb(0x20, 0x20);
}


void TSpawnWindow::spawn()
{
    // first call our new suspend()
    TProgram::application->suspend();

    // ... then save old INT 29h vector and replace it with our own
    void interrupt (*OldInt29)(...) = getvect(INT_NUMBER);
    setvect(INT_NUMBER, (void interrupt (*)(...))NewInt29);

    // ... define a conio window that matches our TSpawnWindow's client area,
    // activate direct video and set the appropiate color
    window(origin.x+3, origin.y+3, origin.x + size.x -2, origin.y + size.y);
    directvideo = 1;
    textattr(getColor(2));

    // ... call some DOS commands (spawn() or system())
    printf("Executing %s\n", DOS_CMD);
    system(DOS_CMD);
    system("pause");

    // ... and restore our old environment
    setvect(INT_NUMBER, OldInt29);
    TProgram::application->resume();

    // finally close our TSpawnWindow
    close();
}

void TSpawnWindow::handleEvent(TEvent& event)
{
  if( event.what == evBroadcast && event.message.command == cmSpawn )
  {
    spawn();
    clearEvent(event);
  } else
    TWindow::handleEvent(event);
}

int main()
{
    TVApp tvApp;
    tvApp.run();
    return 0;
}
